filetype
" <- @keyword

filetype on
" <- @keyword
"        ^ @keyword

filet detect
" <- @keyword
"     ^ @keyword

filet plugin off
" <- @keyword
"     ^ @keyword

filet indent off
" <- @keyword
"     ^ @keyword
